/*
 *	Control Program for The Magic Time Travel Box
 *
 *	(c) 2009 Martin Mares <mj@ucw.cz>
 */

/*
 *	Pin assignment (component side view; "*" marks inverted signals)
 *
 *				+-------------------+
 *		TPIC serial in	| PB0		PA0 |	tap (analog)
 *		TPIC serial clk	| PB1		PA1 |	steam BNC (analog)
 *		TPIC latch clk	| PB2		PA2 |	antenna BNC (analog)
 *		blue LED *	| PB3/OC0	PA3 |	solar panel (analog)
 *		red lock	| PB4		PA4 |	red button *
 *		MOSI (prog)	| PB5		PA5 |	black button *
 *		MISO (prog)	| PB6		PA6 |	left sw *
 *		SCK (prog)	| PB7		PA7 |	right sw *
 *				| RESET*       AREF |	+5V
 *				| VCC		GND |
 *				| GND	       AVCC |	+5V
 *				| XTAL2		PC7 |
 *				| XTAL1		PC6 |	display D7
 *		UART RXD	| PD0		PC5 |	display D6
 *		UART TXD	| PD1		PC4 |	display D5
 *		big red sw *	| PD2		PC3 |	display D4
 *		magnetic sw *	| PD3		PC2 |	display E (clock)
 *		left meter	| PD4/OC1B	PC1 |	display R/W (0=write, 1=read)
 *		bell sw *	| PD5		PC0 |	display R/S (0=control, 1=data)
 *		dial		| PD6	    OC2/PD7 |	right meter
 *				+-------------------+
 *
 *	TPIC output assignment:
 *
 *		D0	big red LED
 *		D1	big yellow LED
 *		D2	big green LED
 *		D3	green LED
 *		D4	green lamp
 *		D5	relay 2
 *		D6	relay 1 -> bell
 *		D7	orange LED
 */

#define F_CPU 1000000L

#include <avr/io.h>
#include <avr/pgmspace.h>
#include <avr/interrupt.h>
#include <avr/sleep.h>
#include <util/delay.h>

typedef uint8_t byte;

#define B(x) (1U<<(x))
#define SET(p,x) p |= B(x)
#define CLR(p,x) p &= ~B(x)

#define usleep _delay_us

static void msleep(byte millisec)
{
	while (millisec) {
		_delay_ms(1);
		millisec--;
	}
}

static void tpic_init(void)
{
	// Shift register: PB0=SIN, PB1=SRCK, PB2=RCK
	DDRB |= B(PB0) | B(PB1) | B(PB2);
}

static byte tpic_last;

static void tpic_write(byte x)
{
	tpic_last = x;
	for (byte mask=0x80; mask; mask >>= 1) {
		if (x & mask)
			SET(PORTB, PB0);
		else
			CLR(PORTB, PB0);
		SET(PORTB, PB1);
		CLR(PORTB, PB1);
	}
	SET(PORTB, PB2);
	CLR(PORTB, PB2);
}

static void uart_init(void)
{
#define UBR (F_CPU/8/9600-1)
	UBRRH = (byte)(UBR >> 8);
	UBRRL = (byte) UBR;
	UCSRA = B(U2X);
	UCSRB = B(RXEN) | B(TXEN);
	UCSRC = B(URSEL) | B(UCSZ0) | B(UCSZ1);
}

static void uart_putc(byte c)
{
	while (!(UCSRA & B(UDRE)))
		;
	UDR = c;
}

static void uart_putx4(byte c)
{
	if (c < 10)
		uart_putc(c + '0');
	else
		uart_putc(c + 'A' - 10);
}

static void uart_putx(byte c)
{
	uart_putx4(c >> 4);
	uart_putx4(c & 0xf);
}

static void uart_puts_p(prog_char *s)
{
	char ch;
	while (ch = pgm_read_byte(s++))
		uart_putc(ch);
}

static void disp_pull(void)
{
	usleep(1);
	SET(PORTC, PC2);
	usleep(1);
	CLR(PORTC, PC2);
	usleep(1);
}

static byte disp_read_nibble_raw(void)
{
	SET(PORTC, PC2);
	usleep(1);
	byte x = (PINC >> PC3) & 0xf;
	CLR(PORTC, PC2);
	usleep(1);
	return x;
}

static void disp_wait(void)
{
#if 1
	PORTC &= ~(0x0f << PC3);	// switch off pull-ups
	DDRC &= ~(0x0f << PC3);
	CLR(PORTC, PC0);
	SET(PORTC, PC1);
	usleep(2);
	byte stat;
	do {
		stat = disp_read_nibble_raw();
		disp_read_nibble_raw();
	} while (stat & 8);
	CLR(PORTC, PC1);
	DDRC |= 0x0f << PC3;
#else
	msleep(100);
#endif
}

static void disp_write_ctrl_nibble(byte x)
{
	PORTC &= ~((0x0f << PC3) | B(PC0) | B(PC1));	// unset data bus, R/S and R/W
	PORTC |= x << PC3;				// set nibble
	disp_pull();
}

static void disp_write_ctrl(byte x)
{
	disp_write_ctrl_nibble((x >> 4) & 0x0f);
	disp_write_ctrl_nibble(x & 0x0f);
	disp_wait();
}

static void disp_write_data_nibble(byte x)
{
	PORTC &= ~((0x0f << PC3) | B(PC1));		// unset data bus and R/W
	PORTC |= (x << PC3) | B(PC0);			// set nibble and R/S
	disp_pull();
}

static void disp_write_data(byte x)
{
	disp_write_data_nibble((x >> 4) & 0x0f);
	disp_write_data_nibble(x & 0x0f);
	disp_wait();
}

static void disp_init(void)
{
	DDRC |= B(PC0) | B(PC1) | B(PC2) | B(PC3) | B(PC4) | B(PC5) | B(PC6);
	CLR(PORTC, PC7);

	/*
	 *	A little bit of magic. We need to synchronize the 4-bit protocol,
	 *	so we first try to switch to the 8-bit protocol, which is stateless.
	 */
	disp_write_ctrl_nibble(0x3);
	msleep(2);
	disp_write_ctrl_nibble(0x3);
	msleep(2);
	disp_write_ctrl_nibble(0x3);
	msleep(2);

	/*
	 *	We know that we are in the right phase, so we switch to the
	 *	4-bit mode at once.
	 */
	disp_write_ctrl_nibble(0x2);
	msleep(2);
	disp_write_ctrl_nibble(0x2);
	disp_write_ctrl_nibble(0x8);
	msleep(2);

	disp_write_ctrl(0x0c);		// display on, cursor off, blink off
	disp_write_ctrl(0x01);		// clear
	disp_write_ctrl(0x06);		// entry mode: incremental, no scrolling
}

static void disp_clear(void)
{
	disp_write_ctrl(0x01);
}

static void disp_goto(byte row, byte col)
{
	disp_write_ctrl(0x80 + (row ? 0x40 : 0) + col);
}

static void disp_putc(byte ch)
{
	disp_write_data(ch);
}

static void disp_puts_p(prog_char *s)
{
	char ch;
	while (ch = pgm_read_byte(s++))
		disp_putc(ch);
}

static void disp_putx4(byte c)
{
	if (c < 10)
		disp_putc(c + '0');
	else
		disp_putc(c + 'A' - 10);
}

static void disp_putx(byte c)
{
	disp_putx4(c >> 4);
	disp_putx4(c & 0xf);
}

static void adc_fire(byte ch)
{
	ADMUX = B(ADLAR) | (ch << MUX0);
	ADCSRA |= B(ADSC);
}

static byte adc_read(byte ch)
{
	adc_fire(ch);
	while (ADCSRA & B(ADSC))
		;
	return ADCH;
}

static void flash_led(void)
{
	CLR(PORTB, PB3);
	tpic_write(0x8f);
	msleep(255);
	SET(PORTB, PB3);
	tpic_write(0);
	msleep(255);
}

static void debug_mode(void)
{
	// Calibrating meters: set them to maximum
	DDRD |= B(PD4) | B(PD7);
	PORTD |= B(PD4) | B(PD7);

	while (1) {
		CLR(PORTB, PB3);
		tpic_write(0x10);
		msleep(50);

		SET(PORTB, PB3);
		tpic_write(0x9f);
		msleep(50);

#if 0
		static byte chargen = 'A';
		uart_putc(chargen++);
		if (chargen > 'Z')
			chargen = 'A';
#endif

#if 1
		uart_putc('<');
		uart_putx(PINA & 0xf0);
		disp_putx(PINB & 0x10);
		uart_putx(PIND & 0x6c);
		uart_putc('>');
#endif

		disp_goto(1, 0);
		disp_putx(PINA & 0xf0);
		disp_putx(PINB & 0x10);
		disp_putx(PIND & 0x6c);
		disp_putc(' ');
		disp_putx(adc_read(0));
		disp_putx(adc_read(1));
		disp_putx(adc_read(2));
		disp_putx(adc_read(3));
		disp_putc(' ');

#if 1
		uart_putc('[');
		uart_putx(adc_read(0));
		uart_putx(adc_read(1));
		uart_putx(adc_read(2));
		uart_putx(adc_read(3));
		uart_putc(']');
#endif

#if 0
		static byte cgpos;
		disp_goto(1, cgpos++ & 0x0f);
		disp_putc(chargen);
#endif

		if (UCSRA & B(RXC)) {
			byte x = UDR;
			uart_putc('{');
			uart_putx(x);
			uart_putc('}');
		}
	}
}

static void pwm1_init(void)
{
	// CLK_IO directly (1MHz), WGM set to 8-bit fast PWM
	TCCR1A = B(COM1B1) | B(WGM10);
	TCCR1B = B(CS10) | B(WGM12);
	DDRD |= B(PD4);
}

static void pwm2_init(void)
{
	// CLK_IO directly (1MHz), WGM set to 8-bit fast PWM
	TCCR2 = B(WGM20) | B(WGM21) | B(COM21) | B(CS20);
	DDRD |= B(PD7);
}

static void pwm1_set(byte x)
{
	OCR1B = x;
}

static void pwm2_set(byte x)
{
	OCR2 = x;
}

static void led_toggle(void)
{
	static byte l_state;
	if (l_state)
		SET(PORTB, PB3);
	else
		CLR(PORTB, PB3);
	l_state ^= 1;
}

static volatile byte adc_values[4];
static byte adc_i;

static volatile byte dial_rx;
static volatile byte a_rx, b_rx;

ISR(TIMER0_OVF_vect)
{
	TCNT0 = 256 - 5;	// 1024/5 Hz

	// Read results from the ADC
	if (!(ADCSRA & B(ADSC))) {
		adc_values[adc_i] = ADCH;
		adc_i = (adc_i + 1) % 4;
		adc_fire(adc_i);
	}

	// Debouncing input pins
	static byte db_a[3], db_b[3];
#define DEBOUNCE(ary, new, res) do { \
		ary[0] = ary[1], ary[1] = ary[2], ary[2] = new;				\
		res = (ary[0] & ary[1]) | (ary[1] & ary[2]) | (ary[0] & ary[2]);	\
	} while(0)
	DEBOUNCE(db_a, (PINA & 0xf0) | (PINB >> 4), a_rx);
	DEBOUNCE(db_b, PIND & 0x6c, b_rx);

	// Reading the dial
	static byte dial_debounce = 0x80;
	static byte dial_tick;
	static byte dial_cnt;
	static byte dial_pause;
	static byte dial_order;
#define DIAL_DEBOUNCE 2
	if (PIND & B(PD6))
		dial_debounce++;
	else
		dial_debounce--;
	if (dial_debounce >= 0x80 + DIAL_DEBOUNCE) {
		dial_debounce = 0x80 + DIAL_DEBOUNCE;
		dial_tick = 1;
	} else if (dial_debounce < 0x80 - DIAL_DEBOUNCE) {
		dial_debounce = 0x80 - DIAL_DEBOUNCE;
		if (dial_tick) {
			dial_cnt++;
			dial_tick = 0;
			dial_pause = 0;
		}
		if (dial_cnt) {
			dial_pause++;
			if (dial_pause >= 40) {
				dial_rx = dial_cnt | dial_order;
				dial_cnt = 0;
				dial_order += 0x10;
			}
		}
	}
}

#define UART_FIFO_SIZE 16
static volatile byte uart_fifo[UART_FIFO_SIZE];
static volatile byte uart_fifo_r, uart_fifo_w;

ISR(USART_RXC_vect)
{
	while (UCSRA & B(RXC)) {
		byte x = UDR;
		byte i = (uart_fifo_w + 1) % UART_FIFO_SIZE;
		if (i != uart_fifo_r) {
			uart_fifo[uart_fifo_w] = x;
			uart_fifo_w = i;
		}
	}
}

static byte uart_rb_have(void)
{
	return uart_fifo_r != uart_fifo_w;
}

static byte uart_rb_get(void)
{
	cli();
	byte x = uart_fifo[uart_fifo_r];
	uart_fifo_r = (uart_fifo_r + 1) % UART_FIFO_SIZE;
	sei();
	return x;
}

static void handle_send(void)
{
	if (!(UCSRA & B(UDRE)))
		return;

	static byte send_i;
	static byte send_z;

	if (send_i & 1) {
		uart_putx4(send_z & 0x0f);
	} else {
		switch (send_i / 2) {
			case 0:
				UDR = '<';
				send_i = 2;
				return;
			case 1: send_z = a_rx; break;
			case 2:	send_z = b_rx; break;
			case 3:	send_z = adc_values[0]; break;
			case 4:	send_z = adc_values[1]; break;
			case 5:	send_z = adc_values[2]; break;
			case 6:	send_z = adc_values[3]; break;
			case 7: send_z = dial_rx; break;
			case 8: UDR = '\r';
				send_i += 2;
				return;
			default:
				UDR = '\n';
				send_i = 0;
				return;
		}
		uart_putx4(send_z >> 4);
	}
	send_i++;
}

#define PKT_SIZE 16
static byte pkt_buf[PKT_SIZE];
static byte pkt_i = 0xff;		// Will wait for sync

static byte dc4(byte i)
{
	if (i >= pkt_i)
		return 0;
	byte x = pkt_buf[i];
	if (x >= '0' && x <= '9')
		return x - '0';
	x &= 0xdf;
	if (x >= 'A' && x <= 'F')
		return x - 'A' + 10;
	return 0;
}

static byte dc8(byte i)
{
	return (dc4(i) << 4) | dc4(i+1);
}

static void got_pkt(void)
{
	if (!pkt_i || pkt_i >= PKT_SIZE)
		return;

	byte x;
	switch (pkt_buf[0]) {
		case 'C':
			disp_clear();
			break;
		case 'G':
			disp_goto(dc4(1), dc4(2));
			break;
		case 'W':
			for (byte i=1; i<pkt_i; i++)
				disp_putc(pkt_buf[i]);
			break;
		case 'L':
			pwm1_set(dc8(1));
			break;
		case 'R':
			pwm2_set(dc8(1));
			break;
		case 'B':
			x = dc8(1);
			x = (x & 0x9f) | (tpic_last & ~0x9f);
			if (x != tpic_last)
				tpic_write(x);
			break;
	}
}

static void handle_recv(void)
{
	while (uart_rb_have()) {
		byte x = uart_rb_get();
		if (x == '\r' || x == '\n') {
			got_pkt();
			pkt_i = 0;
		} else if (pkt_i < PKT_SIZE)
			pkt_buf[pkt_i++] = x;
	}
}

int main(void)
{
	// LED at PB3
	SET(DDRB, PB3);

	// Pull-ups for inputs
	PORTA = 0xf0;
	PORTB |= B(PB4);
	PORTD = 0x6c;

	// Initialize ADC
	ADMUX = B(ADLAR) | (0 << MUX0);
	ADCSRA = B(ADEN) | B(ADSC) | B(ADPS0) | B(ADPS1);	// prescale clock 8x

	uart_init();
	tpic_init();

	flash_led();

	disp_init();
	disp_goto(0, 0);
	disp_puts_p(PSTR("Magic Box v-1.-1"));
	disp_goto(1, 0);
	disp_puts_p(PSTR("(c) Bear Service"));

	for (byte i=0; i<3; i++) {
		if (!(PINA & B(PA4)))
			debug_mode();
		flash_led();
	}

	disp_clear();
	uart_puts_p(PSTR("\r\nBOOT\r\n"));

	// Set up timer
	TCCR0 = 5;		// Use CLK_IO prescaled by 1024 => 1024 Hz
	TIMSK |= B(TOIE0);

	// Set up UART interrupts
	UCSRB |= B(RXCIE);

	// Initialize the PWM's
	pwm1_init();
	pwm1_set(0);
	pwm2_init();
	pwm2_set(0);

	sei();
	adc_fire(0);

	// Wait for a moment and hope that all ADC's finish on time
	for (byte i=0; i<5; i++)
		msleep(100);

	for(;;) {
		handle_send();
		handle_recv();
//		sleep_mode();
	}
}
